from typing import Union

import torch
from torch import nn


class DoubleConv(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.conv_op = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        return self.conv_op(x)


class DownSample(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.conv = DoubleConv(in_channels, out_channels)
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)

    def forward(self, x):
        down = self.conv(x)
        p = self.pool(down)

        return down, p


class UpSample(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.up = nn.ConvTranspose2d(in_channels, in_channels // 2, kernel_size=2, stride=2)
        self.conv = DoubleConv(in_channels, out_channels)

    def forward(self, x1, x2):
        x1 = self.up(x1)
        x = torch.cat([x1, x2], 1)
        return self.conv(x)

def multiply_to_int(a:Union[int, float], b:Union[int, float]) -> int:
    return int(round(a*b))

class UNet(nn.Module):
    def __init__(self, in_channels, num_classes, model_scale=1):
        super().__init__()
        self.down_convolution_1 = DownSample(in_channels, multiply_to_int(64, model_scale))
        self.down_convolution_2 = DownSample(multiply_to_int(64, model_scale), multiply_to_int(128, model_scale))
        self.down_convolution_3 = DownSample(multiply_to_int(128, model_scale), multiply_to_int(256, model_scale))
        self.down_convolution_4 = DownSample(multiply_to_int(256, model_scale), multiply_to_int(512, model_scale))

        self.bottle_neck = DoubleConv(multiply_to_int(512, model_scale), multiply_to_int(1024, model_scale))

        self.up_convolution_1 = UpSample(multiply_to_int(1024, model_scale), multiply_to_int(512, model_scale))
        self.up_convolution_2 = UpSample(multiply_to_int(512, model_scale), multiply_to_int(256, model_scale))
        self.up_convolution_3 = UpSample(multiply_to_int(256, model_scale), multiply_to_int(128, model_scale))
        self.up_convolution_4 = UpSample(multiply_to_int(128, model_scale), multiply_to_int(64, model_scale))

        self.out = nn.Conv2d(in_channels=multiply_to_int(64, model_scale), out_channels=num_classes, kernel_size=1)



    def forward(self, x):
        down_1, p1 = self.down_convolution_1(x)
        down_2, p2 = self.down_convolution_2(p1)
        down_3, p3 = self.down_convolution_3(p2)
        down_4, p4 = self.down_convolution_4(p3)

        b = self.bottle_neck(p4)

        up_1 = self.up_convolution_1(b, down_4)
        up_2 = self.up_convolution_2(up_1, down_3)
        up_3 = self.up_convolution_3(up_2, down_2)
        up_4 = self.up_convolution_4(up_3, down_1)


        out = self.out(up_4)
        return out



import logging
from pathlib import Path
from typing import Any, Union

from rioxarray import rioxarray

from segmentmytif.logging_config import log_array

logger = logging.getLogger(__name__)


def read_geotiff(raster_path: Path) -> rioxarray.raster_array:
    raster = rioxarray.open_rasterio(raster_path)
    log_array(raster.data, logger, array_name=str(raster_path))
    return raster


def save_tiff(prediction_raster: rioxarray.raster_array, output_path: Union[Path, str]) -> None:
    prediction_raster.rio.to_raster(output_path)



import argparse
from enum import Enum
from pathlib import Path
from typing import Literal

import numpy as np
import dask.array as da
from numpy import ndarray
from sklearn.ensemble import RandomForestClassifier
import geopandas as gpd
import rioxarray
import dask
import xarray as xr
from sklearn.linear_model import LogisticRegression
from sklearn.neural_network import MLPClassifier
from sklearn.svm import SVC
from xgboost import XGBClassifier
import logging

from segmentmytif.features import get_features, FeatureType, DEFAULT_CHUNK_OVERLAP
from segmentmytif.logging_config import setup_logger, log_duration, log_array
from segmentmytif.utils.io import read_geotiff, save_tiff
from segmentmytif.utils.geospatial import get_label_array

logger = logging.getLogger(__name__)
if not logger.handlers:
    logger = setup_logger(__name__)


def read_input_and_labels_and_save_predictions(
    raster_path: Path,
    pos_labels_path: Path,
    neg_labels_path: Path,
    output_path: Path,
    feature_type=FeatureType.IDENTITY,
    features_path: Path = None,
    compute_mode: Literal["normal", "parallel", "safe"] = "normal",
    chunks: dict = None,
    chunk_overlap: int = DEFAULT_CHUNK_OVERLAP,
    logger_root: logging.Logger = None,
    **extractor_kwargs,
) -> None:
    # Hijack the global logger if a logger_root is provided
    # This is designed to be used in QGIS environment
    if logger_root is not None:
        global logger
        logger = logger_root

    logger.info("read_input_and_labels_and_save_predictions called with the following arguments:")
    for k, v in locals().items():
        logger.info(f"{k}: {v}")

    # Set compute mode, and get dask kwargs for reading raster data
    dask_kwargs = _set_compute_mode(compute_mode, chunks)

    # Load raster data
    raster = read_geotiff(raster_path)

    # Ensure the raster has correct crs
    # In QGIS environment, rasterio may not be able to read the crs correctly
    # However, osgeo.gdal, which is in QGIS, can read the crs correctly
    # Therefore, we use osgeo.gdal to read the crs, when osgeo is available
    try:
        from osgeo import gdal

        dataset = gdal.Open(raster_path)
        projection = dataset.GetProjection()
        raster = raster.rio.write_crs(projection, inplace=True)
        logger.info("Used osgeo.gdal to read the crs")
    except ImportError:
        logger.info("Used rioxarray to read the crs")

    # Extract features
    features = get_features(
        raster,
        raster_path,
        feature_type,
        features_path,
        chunk_overlap=chunk_overlap,
        compute_mode=compute_mode,
        **extractor_kwargs,
    )

    # Load vector labels as geodataframes, and align CRS with input data
    pos_gdf = gpd.read_file(pos_labels_path).to_crs(raster.rio.crs)
    neg_gdf = gpd.read_file(neg_labels_path).to_crs(raster.rio.crs)

    # Get label arrays
    labels = get_label_array(features, pos_gdf, neg_gdf, compute_mode=compute_mode)

    # Make predictions
    prediction_map = make_predictions(features.data, labels.data)

    # Use raster as the template and assign data
    prediction_raster = raster.isel(band=0).drop_vars(["band"]).expand_dims(band=prediction_map.shape[0])
    prediction_raster.data = prediction_map

    # Save predictions
    prediction_raster.rio.to_raster(output_path)

    return output_path


def make_predictions(input_data: ndarray, labels: ndarray) -> ndarray:
    """Makes predictions by training a classifier and using it for inference.

    Expects input data with shape of [channels, width, height] and labels of shape [classes, width, height]
        :param input_data: input data with shape of [channels, width, height]
        :param labels: labels with shape [1, width, height]
    :return: probabilities with shape [class_values, width, height]
    """
    with log_duration("Prepare train data", logger):
        train_data, train_labels = prepare_training_data(input_data, labels)

    classifier = get_classifier()

    with log_duration("Train model", logger):
        classifier.fit(train_data, train_labels)

    with log_duration("Make predictions", logger):
        predictions = classifier.predict_proba(input_data.reshape((input_data.shape[0], -1)).transpose())
        prediction_map = predictions.transpose().reshape((predictions.shape[1], *input_data.shape[1:]))
        log_array(prediction_map, logger, array_name="Predictions")

    return prediction_map

class ClassifierType(Enum):
    RANDOM_FOREST = 1
    XGBOOST = 2
    MLP = 3
    SVM = 4
    LOGISTIC_REGRESSION = 5


    @staticmethod
    def from_string(s):
        try:
            return ClassifierType[s]
        except KeyError:
            raise ValueError()

def get_classifier(classifier_type = ClassifierType.RANDOM_FOREST):
    logger.info(f"Using classifier: {classifier_type.name}")
    if classifier_type == ClassifierType.RANDOM_FOREST:
        return RandomForestClassifier(n_estimators=100)
    if classifier_type == ClassifierType.XGBOOST:
        return XGBClassifier(n_estimators=5000)
    if classifier_type == ClassifierType.MLP:
        return MLPClassifier()
    if classifier_type == ClassifierType.SVM:
        return SVC(probability=True)
    if classifier_type == ClassifierType.LOGISTIC_REGRESSION:
        return LogisticRegression()

    raise ValueError(f"Invalid classifier type: {classifier_type}")


def prepare_training_data(input_data, labels):
    """
    Prepares training data for a binary classification task.

    Parameters:
    - input_data: A 3D array-like object (e.g., NumPy or Dask array) where the first dimension represents
      instances, and the last two dimensions represent spatial data.
    - labels: A 3D array-like object (e.g., NumPy or Dask array) where the first dimension is the class index
      (only single-class supported), and the last two dimensions represent spatial labels.

    Process:
    1. Flattens the label array for the first class to a 1D array.
    2. Separates the positive and negative instances from the input data based on the labels.
    3. Computes the number of labeled and unlabeled instances and logs the statistics.
    4. Concatenates positive and negative instances into training data and corresponding labels.

    Returns:
    - train_data: A 2D array where each row is a training instance and each column is a feature.
    - train_labels: A 1D array containing the labels corresponding to the rows in `train_data`.
    """
    # Reshape input data to [n_instances, n_features]
    class1_labels = labels[0]  # Only single class is supported
    flattened = class1_labels.flatten()
    positive_instances = input_data.reshape((input_data.shape[0], -1))[:, flattened == 1].transpose()
    negative_instances = input_data.reshape((input_data.shape[0], -1))[:, flattened == 0].transpose()
    n_labeled = flattened.shape[0]
    n_unlabeled = np.prod(labels.shape[-2:]) - n_labeled
    logger.info(
        f"Dataset contains {n_labeled} ({round(100 * n_labeled / (n_labeled + n_unlabeled), 2)}%) labeled instances of a total of {n_labeled + n_unlabeled} instances."
    )
    # Subsample training data
    sampled_positive_instances = subsample(positive_instances, 10000)
    sampled_negative_instances = subsample(negative_instances, 10000)
    n_sampled_positive = sampled_positive_instances.shape[0]
    n_sampled_negative = sampled_negative_instances.shape[0]
    n_sampled_labeled = n_sampled_negative + n_sampled_positive
    logger.info(
        f"Training on {n_sampled_positive} ({round(100 * n_sampled_positive / n_sampled_labeled, 2)}%) positive labels and {n_sampled_negative} ({round(100 * n_sampled_negative / n_sampled_labeled, 2)}%) negative labels "
    )
    # Shuffle training data
    total_sample_size = sampled_positive_instances.shape[0] + sampled_negative_instances.shape[0]
    order = np.arange(total_sample_size)
    np.random.shuffle(order)
    train_data = np.concatenate((sampled_positive_instances, sampled_negative_instances))[order]
    train_labels = np.concatenate(
        (
            (np.ones(shape=[sampled_positive_instances.shape[0]])),
            (np.zeros(shape=[sampled_negative_instances.shape[0]])),
        )
    )[order]
    log_array(train_labels, logger, array_name="Train labels")
    log_array(train_data, logger, array_name="Train data")

    return train_data, train_labels


def subsample(instances, sample_size):
    if isinstance(instances, da.Array):
        instances.compute_chunk_sizes()
    n_instances = instances.shape[0]
    indices = np.arange(n_instances)
    sample_indices = np.random.choice(indices, size=min(sample_size, n_instances), replace=False)
    return instances[sample_indices]


def _set_compute_mode(compute_mode: Literal["normal", "parallel", "safe"], chunks: dict) -> dict:
    """Set compute mode for read_input_and_labels_and_save_predictions."""

    # Set default chunks
    if chunks is None:
        chunks = {"band": 1, "x": 1024, "y": 1024}

    # Default dask kwargs
    dask_kwargs = {}

    match compute_mode:
        case "normal":
            return dask_kwargs
        case "parallel":
            dask.config.set(scheduler="threads")
            dask_kwargs["chunks"] = chunks
        case "safe":
            dask.config.set(scheduler="synchronous")
            dask_kwargs["chunks"] = chunks
        case _:
            msg = f"Invalid compute mode: {compute_mode}"
            raise ValueError(msg)
    return dask_kwargs


def parse_args():
    parser = argparse.ArgumentParser(
        description="Classify pixels in the input image using a model trained on the labels."
    )

    parser.add_argument("-i", "--input", type=Path, help="Path to the input TIFF file")
    parser.add_argument("-lp", "--pos_labels", type=Path, help="Path to the positive training labels file, shp or gpkg")
    parser.add_argument("-ln", "--neg_labels", type=Path, help="Path to the negative training labels file, shp or gpkg")
    parser.add_argument("-p", "--predictions", type=Path, help="Path to the predictions output TIFF file")
    parser.add_argument(
        "-f",
        "--feature_type",
        type=FeatureType.from_string,
        choices=list(FeatureType),
        default=FeatureType.FLAIR,
        help=f"Type of feature being used. Default: {FeatureType.FLAIR.name}",
    )
    parser.add_argument(
        "-m",
        "--compute_mode",
        type=str,
        choices=["normal", "parallel", "safe"],
        default="normal",
        help="Mode for reading the input raster data.",
    )
    parser.add_argument(
        "-c",
        "--chunks",
        type=dict,
        default={"band": 1, "x": 1024, "y": 1024},
        help="Chunk size for reading the input raster data.",
    )
    parser.add_argument(
        "-o",
        "--chunk_overlap",
        type=int,
        default=DEFAULT_CHUNK_OVERLAP,
        help="Chunk overlap size for reading the input raster data.",
    )

    args = parser.parse_args()

    # Validate arguments
    if not args.input.exists() or not args.input.is_file():
        parser.error(f"The input file {args.input} does not exist or is not a file.")

    return args


if __name__ == "__main__":
    args = parse_args()
    input_path = args.input
    pos_labels_path = args.pos_labels
    neg_labels_path = args.neg_labels
    predictions_path = args.predictions
    feature_type = args.feature_type

    read_input_and_labels_and_save_predictions(
        input_path, pos_labels_path, neg_labels_path, predictions_path, feature_type=feature_type
    )



import shutil
from pathlib import Path

import pytest

from segmentmytif.utils.monochromize import monochromize_image, monochromize_folder
from .test_cases import test_case1210
from .utils import TEST_DATA_FOLDER


@pytest.mark.xfail(reason="This tests errors with 'UnicodeDecodeError: 'utf-8' codec can't decode byte 0xba in position 18: invalid start byte'. I'm not sure why.")
def test_monochromize(tmpdir):
    input_image = TEST_DATA_FOLDER / test_case1210.image_filename
    monochromize_image(input_image, tmpdir)
    for channel in range(4):
        assert (tmpdir / f"test_image_{channel}.tif").exists()


def test_monochromize_folder_structure_maintained(tmpdir):
    """Tests if the folder structure is maintained from the source to the target folder.

    - source
      - subdir
        - test_image_1210x718.tif
    """
    input_image = TEST_DATA_FOLDER / test_case1210.image_filename
    source_dir = Path(tmpdir) / "source"
    subdir = "subdir"
    (source_dir / subdir).mkdir(parents=True)
    shutil.copyfile(input_image, source_dir / subdir / input_image.name)
    target_dir = Path(tmpdir) / "target"
    target_dir.mkdir()

    monochromize_folder(source_dir, target_dir)

    for channel in range(4):
        tiff_path = (target_dir / subdir / f"{input_image.stem}_0{input_image.suffix}")
        assert tiff_path.exists(), f"tif file should have been created at {str(tiff_path)} for channel {channel}"



"""Geosptial related utilities."""

from typing import Literal
import geopandas as gpd
import numpy as np
import xarray as xr


def get_label_array(
    input_data: xr.DataArray,
    geom_pos: gpd.GeoSeries,
    geom_neg: gpd.GeoSeries,
    compute_mode: Literal["normal", "parallel", "safe"] = "normal",
) -> xr.DataArray:
    """Generate a label array for binary classification from pos/neg geometries.

    The function overlays pos/neg geometries on a template raster,
    then fills the overlapped area with the label value,
    leaving the rest of the raster with -1.

    There are three values filled:
    - 1: positive label
    - 0: negative label
    - -1: unclassified

    One needs to call geom_to_label_array at least twice to get a
    complete label array for a binary classification task.

    There are two modes of execution:
    - normal: Assumes the label array fits in memory
    - parallel or safe: Assumes the label array is too large to fit in memory, do block processing.
        The difference between parallel and safe is safe only uses single thread, which will be
        configured in dask schedular.

    :param xr.DataArray input_data: input features as a template raster
    :param gpd.GeoSeries geom: Geometry of the label
    :param Literal["normal", "parallel", "safe"] mode: Mode of execution, defaults to "normal"
    :return xr.DataArray: Generated label array
    """
    # If the template raster has a "band" dimension
    # Select the first band, since label will be 2D
    if "band" in input_data.dims:
        input_template = input_data.isel(band=0).drop_vars("band")


    positive_labels = _geom_to_label_array(input_template, geom_pos, 1, compute_mode=compute_mode)
    negative_labels = _geom_to_label_array(input_template, geom_neg, 0, compute_mode=compute_mode)

    labels = -(positive_labels * negative_labels)  # Combine positive and negative labels

    # repeat to match the input_template
    return labels.expand_dims(dim={"band": input_data.sizes["band"]})


def _geom_to_label_array(
    input_template: xr.DataArray,
    geom: gpd.GeoSeries,
    value: Literal[0, 1],
    compute_mode: Literal["normal", "parallel", "safe"] = "normal",
) -> xr.DataArray:
    """Generate a label array from a geometry."""
    # Make a template from the shape of the template raster
    # Fill it with the label value
    labels_template = xr.full_like(input_template, fill_value=value, dtype=np.int32)

    # Set the nodata value to -1 indicating other classes
    labels_template = labels_template.rio.write_nodata(-1)

    match compute_mode:
        # Assumning labels_template can fit in memory
        case "normal":
            label_array = labels_template.rio.clip(geom.geometry, drop=False)
        # Block processing for large rasters
        case "parallel" | "safe":
            label_array = xr.map_blocks(
                lambda raster, geom: raster.rio.clip(geom, drop=False),
                labels_template,
                args=(geom.geometry,),
                template=labels_template,
            )
        case _:
            msg = f"Mode {compute_mode} not supported"
            raise ValueError(msg)

    return label_array



"""Documentation about segmentmytif."""

import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())

__author__ = "Christiaan Meijer"
__email__ = "c.meijer@esciencecenter.nl"
__version__ = "0.3.0"



import numpy as np
import torch
from torchinfo import summary

from segmentmytif.utils.datasets import normalize_single_band_to_tensor
from segmentmytif.utils.models import UNet
from tests.utils import TEST_DATA_FOLDER


class TestUNet:
    def test_initialize_unet_with_valid_parameters(self):
        """Initialize UNet with valid in_channels and num_classes"""
        in_channels = 3
        num_classes = 19

        model = UNet(in_channels, num_classes)

        assert isinstance(model, UNet)
        assert model.out.out_channels == num_classes

    def test_forward_with_minimum_valid_dimensions(self):
        """Handle input tensor with minimum valid dimensions"""
        in_channels = 3
        num_classes = 19
        width = height = 64
        input_tensor = torch.randn(1, in_channels, width, height)  # Minimum size for U-Net to work

        model = UNet(in_channels, num_classes)
        output = model(input_tensor)

        assert output.shape == (1, num_classes, width, height)

    def test_summary(self):
        """Summary of model is exactly as tested."""
        in_channels = 3
        num_classes = 19
        width = height = 64
        input_tensor = torch.randn(1, in_channels, width, height)  # Minimum size for U-Net to work
        expected_sum = load_summary('test_model_summary.json')

        model = UNet(in_channels, num_classes)
        sum = get_summary(model, input_tensor)

        assert sum == expected_sum

    def test_summary_model_scaled_down(self):
        """Summary of model is exactly as tested."""
        model_scale = 0.5
        in_channels = 3
        num_classes = 19
        width = height = 64
        input_tensor = torch.randn(1, in_channels, width, height)  # Minimum size for U-Net to work
        expected_sum = load_summary('test_model_summary_half.json')

        model = UNet(in_channels, num_classes, model_scale=model_scale)
        sum = get_summary(model, input_tensor)

        assert sum == expected_sum

    def test_summary_model_scaled_up(self):
        """Summary of up-scaled model is exactly as tested."""
        model_scale = 2
        in_channels = 3
        num_classes = 19
        width = height = 64
        input_tensor = torch.randn(1, in_channels, width, height)  # Minimum size for U-Net to work
        expected_sum = load_summary('test_model_summary_double.json')

        model = UNet(in_channels, num_classes, model_scale=model_scale)
        sum = get_summary(model, input_tensor)

        assert sum == expected_sum


def load_summary(file_name):
    expected_sum = (TEST_DATA_FOLDER / file_name).read_text(encoding='utf-8')
    return normalize_model_summary(expected_sum)


def get_summary(model, input_tensor):
    return normalize_model_summary(summary(model, input_data=input_tensor, verbose=0).__repr__())


def normalize_model_summary(text):
    return (text
            .replace("=", "")
            .replace("(G)", "(Units.GIGABYTES)")
            .replace("(M)", "(Units.MEGABYTES)")
            )


def test_normalize_single_band_to_tensor_uint8_array():
    """Test normalization with a regular array having non-zero mean and std."""
    # Regular array with non-zero mean and std
    test_array = np.random.randint(0, 255, (100, 100), dtype=np.uint8)

    normalized = normalize_single_band_to_tensor(test_array)

    expected_shape = test_array.shape
    verify_normalized_ouput(normalized, expected_shape)


def test_normalize_single_band_to_tensor_regular_array():
    """Test normalization with a regular array having non-zero mean and std."""
    # Regular array with non-zero mean and std
    test_array = np.random.rand(100, 100) * 10 + 5  # Values between 5 and 15

    normalized = normalize_single_band_to_tensor(test_array)

    expected_shape = test_array.shape
    verify_normalized_ouput(normalized, expected_shape)


def verify_normalized_ouput(normalized, expected_shape, expected_std=1.0):
    # Check that the output is a torch tensor with correct dimensions
    assert isinstance(normalized, torch.Tensor)
    assert normalized.shape == expected_shape
    # Check normalization (should have mean ~0 and std ~1)
    normalized_numpy = normalized.numpy()
    assert abs(normalized_numpy.mean()) < 1e-6  # Close to zero
    assert abs(normalized_numpy.std() - expected_std) < 1e-6  # Close to one


def test_normalize_single_band_to_tensor_zero_std_array():
    """Test normalization with an array having zero standard deviation."""
    # Array with zero standard deviation
    zero_std_array = np.ones((50, 50)) * 7  # All values are 7
    zero_std_normalized = normalize_single_band_to_tensor(zero_std_array, debug_note="zero std test")

    verify_normalized_ouput(zero_std_normalized, zero_std_array.shape, expected_std=0.0)



from pathlib import Path

import numpy as np

from segmentmytif.utils.io import read_geotiff, save_tiff


def generate_train_labels(data: np.ndarray) -> np.ndarray:
    train_input = np.zeros((1, *data.shape[1:]), dtype=np.int32) - 1
    train_input[0, 540:, 1090:] = 1  # water
    train_input[0, 678:, 516:560] = 0  # grass
    return train_input

def main():
    input_image_path = Path("tests/test_data/test_image.tif")
    labels_path = Path("tests/test_data/test_image_labels.tif")
    data, profile = read_geotiff(input_image_path)
    save_tiff(generate_train_labels(data), labels_path, profile)


if __name__ == '__main__':
    main()


import argparse
from pathlib import Path

from tqdm import tqdm

from segmentmytif.utils.io import read_geotiff, save_tiff


def parse_args():
    parser = argparse.ArgumentParser(description="Process input and output TIFF files.")

    parser.add_argument('-i', '--input_folder', type=Path, help='Path to a folder of input tiff files', required=True)
    parser.add_argument('-o', '--output_folder', type=Path, help='Path to the output folder', required=True)

    args = parser.parse_args()

    if not args.input_folder.exists():
        parser.error(f"The input folder {args.input} does not exist.")

    return args


def monochromize_image(input_file_path: Path, output_folder_path: Path):
    data = read_geotiff(input_file_path)
    for i_channel in range(data.shape[0]):
        output_file_name = f"{input_file_path.stem}_{i_channel}{input_file_path.suffix}"
        channel = data[i_channel:i_channel + 1]
        save_tiff(channel, output_folder_path / output_file_name)


def monochromize_folder(input_folder: Path, output_folder: Path):
    output_folder.mkdir(parents=True, exist_ok=True)
    for input_file in tqdm(list(input_folder.rglob('*.tif')), desc="Processing images"):
        relative_path = input_file.relative_to(input_folder)
        if input_file.is_file():
            path = output_folder / relative_path.parent
            path.mkdir(parents=True, exist_ok=True)
            monochromize_image(input_file, path)


if __name__ == "__main__":
    args = parse_args()
    input_folder = args.input_folder
    output_folder = args.output_folder
    monochromize_folder(input_folder, output_folder)



import shutil
from pathlib import Path

import dask.array as da
import numpy as np
import pandas as pd
import pytest
import rasterio
import rioxarray
from scipy.spatial.distance import dice
import geopandas as gpd
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import xarray as xr
from segmentmytif.features import FeatureType, get_features_path, get_features
from segmentmytif.main import read_input_and_labels_and_save_predictions, prepare_training_data, make_predictions
from segmentmytif.utils.geospatial import get_label_array
from segmentmytif.utils.io import save_tiff
from .test_cases import TestCase, test_case1210, test_case512
from .utils import TEST_DATA_FOLDER


@pytest.mark.parametrize("test_case, feature_type, model_scale, dice_similarity_threshold, compute_mode",
                         [
                             pytest.param(test_case1210, FeatureType.IDENTITY, None, None, "normal", marks=pytest.mark.slow),
                             pytest.param(test_case1210, FeatureType.FLAIR, 0.125, None, "normal"), # also slow, but necessary to test on each run
                             pytest.param(test_case512, FeatureType.IDENTITY, None, 0.90, "normal", marks=pytest.mark.slow),
                             pytest.param(test_case512, FeatureType.FLAIR, 0.125, 0.82, "normal", marks=pytest.mark.slow),
                             pytest.param(test_case512, FeatureType.FLAIR, 1.0, 0.95, "normal", marks=pytest.mark.slow),
                             pytest.param(test_case512, FeatureType.FLAIR, 1.0, 0.95, "parallel", marks=pytest.mark.slow),
                             pytest.param(test_case512, FeatureType.FLAIR, 1.0, 0.95, "safe", marks=pytest.mark.slow),
                             pytest.param(test_case1210, FeatureType.IDENTITY, None, None, "parallel", marks=pytest.mark.slow),
                             pytest.param(test_case1210, FeatureType.FLAIR, 0.125, None, "parallel", marks=pytest.mark.slow),
                             pytest.param(test_case1210, FeatureType.IDENTITY, None, None, "safe", marks=pytest.mark.slow),
                             pytest.param(test_case1210, FeatureType.FLAIR, 0.125, None, "safe", marks=pytest.mark.slow),
                         ], ids=lambda e : str(e))
def test_integration(tmpdir, test_case: TestCase, feature_type, model_scale, dice_similarity_threshold, compute_mode):
    tmpdir = Path(tmpdir)

    input_path = copy_file_and_get_new_path(test_case.image_filename, tmpdir)
    labels_pos_path = copy_file_and_get_new_path(test_case.labels_pos_filename, tmpdir)
    labels_neg_path = copy_file_and_get_new_path(test_case.labels_neg_filename, tmpdir)
    predictions_path = Path(tmpdir) / f"{test_case.image_filename}_predictions_{str(feature_type)}_model_{model_scale}.tif"

    read_input_and_labels_and_save_predictions(
        input_path, labels_pos_path, labels_neg_path, predictions_path, feature_type=feature_type, model_scale=model_scale
    )

    assert predictions_path.exists()

    # Check DICE similarity if threshold is provided
    if dice_similarity_threshold is None and test_case.ground_truth_filename is None:
        return

    truth = rioxarray.open_rasterio(TEST_DATA_FOLDER / "test_image_512x512_out_ground_truth.tif").astype(np.int16)
    predictions = rioxarray.open_rasterio(predictions_path).astype(np.float64)
    dice_similarity = 1 - dice(truth.data.flatten() > 0.5, predictions.data.flatten() > 0.999)
    print(f"DICE similarity index: {dice_similarity}")
    for t in range(0,100):
        dice_similarity = 1 - dice(truth.data.flatten() > 0.5, predictions.data.flatten() > t/100)
        print(f"DICE similarity index ({t/100}): {dice_similarity}")
    assert dice_similarity > dice_similarity_threshold


def copy_file_and_get_new_path(test_image, tmpdir):
    input_path = Path(tmpdir) / test_image
    shutil.copy(TEST_DATA_FOLDER / test_image, input_path)
    return input_path


@pytest.mark.parametrize(
    "input_path, feature_type, expected_path",
    [
        ("input.tiff", FeatureType.FLAIR, "input_FLAIR.tiff"),
        ("../path/to/input.tiff", FeatureType.FLAIR, "../path/to/input_FLAIR.tiff"),
        ("../path/to/input.tiff", FeatureType.IDENTITY, "../path/to/input.tiff"),
    ],
)
def test_get_features_path(input_path, feature_type, expected_path):
    features_path = get_features_path(Path(input_path), feature_type)
    assert features_path == Path(expected_path)


@pytest.mark.parametrize("array_type", ["numpy", "dask"])
def test_prepare_training_data(array_type):
    random = np.random.default_rng(0)
    length = 200
    random_data = random.integers(low=0, high=256, size=(5, length, length))
    labels = random.choice([0, 1, 2], size=(1, length, length), replace=True)
    if array_type == "numpy":
        input_data = random_data
    elif array_type == "dask":
        input_data = da.from_array(random_data)

    prepare_training_data(input_data, labels)


def train_predict_score(features, raster, tmpdir, use_case):
    pos_gdf = gpd.read_file(TEST_DATA_FOLDER / use_case.labels_pos_filename).to_crs(raster.rio.crs)
    neg_gdf = gpd.read_file(TEST_DATA_FOLDER / use_case.labels_neg_filename).to_crs(raster.rio.crs)
    predictions_path = Path(tmpdir) / "predictions.tif"
    # Get label arrays
    labels = get_label_array(features, pos_gdf, neg_gdf, compute_mode="normal")
    # Make predictions
    prediction_map = make_predictions(features.data, labels.data)
    # Use raster as the template and assign data
    prediction_raster = raster.isel(band=0).drop_vars(["band"]).expand_dims(band=prediction_map.shape[0])
    prediction_raster.data = prediction_map
    # Save predictions
    prediction_raster.rio.to_raster(predictions_path)
    truth = rioxarray.open_rasterio(TEST_DATA_FOLDER / use_case.ground_truth_filename).astype(np.int16)
    predictions = rioxarray.open_rasterio(predictions_path).astype(np.int16)
    dice_similarity = 1 - dice(truth.data.flatten(), predictions.data.flatten())
    return dice_similarity



from dataclasses import dataclass
from typing import Optional


@dataclass
class TestCase:
    __test__ = False  # This is to prevent pytest from running this class as a test case
    image_filename: str
    labels_pos_filename: str
    labels_neg_filename: str
    ground_truth_filename: Optional[str]

    def __repr__(self):
        return self.image_filename

    def __str__(self):
        return self.image_filename


test_case1210=TestCase(
    image_filename="test_image_1210x718.tif",
    labels_pos_filename="test_image_labels_positive_1210x718.gpkg",
    labels_neg_filename="test_image_labels_negative_1210x718.gpkg",
    ground_truth_filename=None,
)
test_case512=TestCase(
    image_filename="test_image_512x512.tif",
    labels_pos_filename="test_image_labels_positive_512x512.gpkg",
    labels_neg_filename="test_image_labels_negative_512x512.gpkg",
    ground_truth_filename="test_image_512x512_out_ground_truth.tif",
)



import numpy as np
import pytest
import torch

from segmentmytif.features import (
    get_features,
    extract_features,
    FeatureType,
    NUM_FLAIR_CLASSES,
    get_flair_model_file_name,
    load_model, pad, calculate_pad_sizes_1d, unpad,
)
from segmentmytif.utils.models import UNet


class TestExtractFeatures:
    def test_get_identity_features(self):
        input_data = np.array(get_generated_multiband_image())
        result = get_features(input_data, None, FeatureType.IDENTITY, None)
        assert np.array_equal(result.data, input_data.data)

    @pytest.mark.parametrize(
        ["n_bands", "width", "height"],
        [
            (3, 1, 1),  # too small to be processed by the model, requires padding
            (3, 8, 8),  # too small to be processed by the model, requires padding
            (3, 16, 16),  # smallest size that can natively be processed by the model
            (3, 61, 39),  # not divisible by 16 so requires padding in both directions
            (3, 64, 48),  # smallest dimensions, > line above, that don't require padding
            (1, 512, 512),  # size of the model's training data (easiest case)
            (3, 1210, 718),  # not divisible by 16 so requires padding in both directions
        ],
    )
    def test_extract_flair_features(self, n_bands, width, height):
        input_data = np.array(get_generated_multiband_image(n_bands=n_bands, width=width, height=height))
        result = extract_features(input_data, FeatureType.FLAIR, model_scale=0.125)
        assert np.array_equal(result.shape, [n_bands * NUM_FLAIR_CLASSES] + list(input_data.shape[1:]))

    def test_extract_features_unsupported_type(self):
        input_data = np.array([[1, 2], [3, 4]])
        with pytest.raises(KeyError):
            extract_features(input_data, "UNSUPPORTED_TYPE")


def get_generated_multiband_image(n_bands=3, width=512, height=512):
    return np.random.random(size=[n_bands, width, height])


@pytest.mark.parametrize(
    ["model_scale", "file_name"],
    [
        (1.0, "flair_toy_ep15_scale1_0.pth"),
        (0.5, "flair_toy_ep15_scale0_5.pth"),
        (0.25, "flair_toy_ep15_scale0_25.pth"),
        (0.125, "flair_toy_ep15_scale0_125.pth"),
    ],
)
def test_get_flair_model_file_name_with_valid_scales(model_scale, file_name):
    assert get_flair_model_file_name(model_scale) == file_name


@pytest.mark.parametrize(["model_scale"], [(2,), (0.001,)])
def test_get_flair_model_file_name_with_invalid_scales(model_scale):
    with pytest.raises(ValueError):
        get_flair_model_file_name(model_scale)


@pytest.mark.downloader
@pytest.mark.parametrize(["model_scale", "source"], [
    (0.125, "Huggingface"),
    (0.25, "Huggingface"),
    (0.5, "Huggingface"),
    (1.0, "Huggingface"),
    (1.0, "Surfdrive"),
    (1.0, None),
])
def test_load_model_downloads(model_scale, source, tmpdir):
    sources = (source,) if source is not None else None
    model, device = load_model(model_scale, models_dir=tmpdir, sources=sources)
    assert model is not None
    assert type(model) == UNet



def test_pad():
    """Test padding of input band to ensure dimensions are divisible by 16."""
    input_band = torch.tensor(np.array([[1, 2, 3], [4, 5, 6]])[None, None, :, :], dtype=torch.float32)
    expected_output = torch.tensor(
        np.array([
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 2, 3, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 4, 5, 6, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        ])[None, None, :, :], dtype=torch.float32
    )

    padded_band = pad(input_band, band_name="test_band")

    assert torch.equal(padded_band, expected_output)


@pytest.mark.parametrize(["input_size"], [((1, 1, 1, 2),), ((1, 1, 17, 23),), ((1, 1, 123, 196),)])
def test_pad_unpad(input_size):
    rng = np.random.default_rng(0)
    input_band = torch.tensor(rng.random(input_size), dtype=torch.float32)

    padded_band = pad(input_band, band_name=f"{input_size}")
    unpadded_band = unpad(padded_band, input_size)

    assert torch.equal(input_band, unpadded_band)


@pytest.mark.parametrize("dim_size, expected_pad_sizes", [
    (1, (7, 8)),
    (4, (6, 6)),
    (15, (0, 1)),
    (16, (0, 0)),
    (17, (7, 8)),
    (31, (0, 1)),
    (32, (0, 0))
])
def test_calculate_pad_size_1d(dim_size, expected_pad_sizes):
    """Test calculation of padding sizes to make a dimension size divisible by 16."""
    pad_sizes = calculate_pad_sizes_1d(dim_size)
    assert pad_sizes == expected_pad_sizes



import logging
import sys
import time
from contextlib import contextmanager
from pathlib import Path
from logging import handlers

import numpy as np
import pandas as pd


def setup_logger(name: str = None, stdout_level=logging.INFO):
    """
    Setup a logger with a specific name and logging level.
    """

    path = Path('log')
    path.mkdir(exist_ok=True, parents=True)

    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")

    # Create handlers
    debug_fh = logging.handlers.RotatingFileHandler(path / 'debug.log', maxBytes=10 * 1024 * 1024, backupCount=5)
    debug_fh.setLevel(logging.DEBUG)
    debug_fh.setFormatter(formatter)
    info_fh = logging.handlers.RotatingFileHandler(path / 'info.log')
    info_fh.setLevel(logging.INFO)
    info_fh.setFormatter(formatter)
    stdout_sh = logging.StreamHandler(sys.stdout)
    stdout_sh.setLevel(stdout_level)
    stdout_sh.setFormatter(formatter)

    # Configure logger and add handlers
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)
    root_logger.addHandler(debug_fh)
    root_logger.addHandler(info_fh)
    root_logger.addHandler(stdout_sh)

    return root_logger


@contextmanager
def log_duration(task_name: str, logger: logging.Logger):
    """
    Log duration of a task.
    """
    logger.info(f"{task_name} started")
    start_time = time.perf_counter()
    yield
    duration = time.perf_counter() - start_time
    logger.info(f"{task_name} finished in {duration:.4f} seconds")


def log_array(data: np.ndarray, logger, array_name:str="array") -> None:
    logger.debug(f"{array_name} shape: {data.shape}")



import logging
from enum import Enum
from pathlib import Path
from typing import Literal

import dask.array as da
import numpy as np
import rioxarray
import torch
import wget
import xarray as xr
from dask.array.core import Array
from huggingface_hub import hf_hub_download
from numpy import ndarray

from segmentmytif.logging_config import log_duration, log_array
from segmentmytif.utils.datasets import normalize_single_band_to_tensor
from segmentmytif.utils.models import UNet

NUM_FLAIR_CLASSES = 19
DEFAULT_CHUNK_OVERLAP = 25  # Default chunk overlap size for feature extraction
logger = logging.getLogger(__name__)


class FeatureType(Enum):
    IDENTITY = 1
    FLAIR = 2

    @staticmethod
    def from_string(s):
        try:
            return FeatureType[s]
        except KeyError:
            raise ValueError()


def get_features(
    raster: xr.DataArray,
    raster_path: Path,
    feature_type: FeatureType,
    features_path: Path,
    chunk_overlap: int = DEFAULT_CHUNK_OVERLAP,
    compute_mode: Literal["normal", "parallel", "safe"] = "normal",
    **extractor_kwargs,
):
    """Extract features from the input data, or load them from disk if they have already been extracted.

    :param raster: raster data as xr.DataArray, read by rioxarray
    :param raster_path: path to the raster tiff file
    :param feature_type: See FeatureType enum for options.
    :param features_path: Path used for caching features
    :param chunk_overlap: Overlap between chunks when chunk-wise processing is enanbled
    :param extractor_kwargs: options for the feature extractor
    :return: extracted features as xr.DataArray
    """
    if feature_type == FeatureType.IDENTITY:
        return raster

    if features_path is None:
        features_path = get_features_path(raster_path, feature_type)
    if not features_path.exists():
        extract_and_save_features(raster, feature_type, chunk_overlap, features_path, extractor_kwargs)

    match compute_mode:
        case "normal":
            loaded_features = rioxarray.open_rasterio(features_path)
        case "parallel" | "safe":
            # Unpack the chunk sizes from the input raster
            chunks = {
                "band": raster.chunksizes["band"][0],
                "x": raster.chunksizes["x"][0],
                "y": raster.chunksizes["y"][0],
            }
            loaded_features = rioxarray.open_rasterio(features_path, chunks=chunks)
    msg = f"Loading {feature_type.name} features (shape {loaded_features.shape}) from {features_path}"
    logger.info(msg)

    return loaded_features


def extract_and_save_features(raster, feature_type, chunk_overlap, features_path, extractor_kwargs):
    msg = (
        f"No existing {feature_type.name} features found at {features_path} "
        f"for input data with shape {raster.data.shape}"
    )
    logger.info(msg)
    with log_duration(f"Extracting {feature_type.name} features", logger):
        features_data = extract_features(raster.data, feature_type, chunk_overlap, **extractor_kwargs)
    log_array(features_data, logger, array_name=f"{feature_type.name} features")
    # Create xarray DataArray with the extracted features
    # Keep the geospatial information from the input raster
    features = raster.isel(band=0).drop_vars(["band"]).expand_dims(band=features_data.shape[0])
    features.data = features_data
    # Save the features to disk
    msg = f"Saving {feature_type.name} features (shape {features_data.shape}) to {features_path}"
    logger.info(msg)
    features.rio.to_raster(features_path)


def extract_features(input_data, feature_type, chunk_overlap=16, **extractor_kwargs):
    extractor = {
        FeatureType.FLAIR: extract_flair_features,
    }[feature_type]

    # If dask array, map feature extraction function to each block
    # The blockwise extraction is only applied to FLAIR features
    # FeatureType.IDENTITY directly returns the input data, thus no need to map_overlap
    if isinstance(input_data, Array):
        # Make template dask array according to the extractor
        if feature_type == FeatureType.FLAIR:
            # If FLAIR, output features has bands * NUM_FLAIR_CLASSES
            meta = da.zeros_like(
                input_data,
                shape=(input_data.shape[0] * NUM_FLAIR_CLASSES, input_data.shape[1], input_data.shape[2]),
            )
        else:
            msg = f"Unsupported feature type: {feature_type}"
            raise ValueError(msg)

        # Feature extraction per block with overlap
        features = input_data.map_overlap(
            extractor,
            **extractor_kwargs,
            depth=(0, chunk_overlap, chunk_overlap),
            boundary="none",
            meta=meta,
        )

        # Since extractor changed the shape of the data
        # We call "compute_chunk_sizes" to align the shape with meta
        features = features.compute_chunk_sizes()

    else:
        features = extractor(input_data, **extractor_kwargs)

    return features


def extract_identity_features(input_data: ndarray) -> ndarray:
    return input_data


def extract_flair_features(input_data: ndarray, model_scale=1.0) -> ndarray:
    """

    :param input_data: Array-like input data as stored in TIFs. Shape: [n_bands, height, width]
    :param model_scale: Scale of the model to use. Must be one of [1.0, 0.5, 0.25, 0.125]
    :return: Features extracted from the input data
    """
    logger.info(f"Using UNet at scale {model_scale}")
    model, device = load_model(model_scale)
    n_bands = input_data.shape[0]

    outputs = []
    for i_band in range(n_bands):
        input_band = normalize_single_band_to_tensor(input_data[i_band:i_band + 1, :, :])[None, :, :, :].float().to(
            device)
        padded_input = pad(input_band, band_name=i_band)
        padded_current_predictions = model(padded_input)
        current_predictions = unpad(padded_current_predictions, input_band.shape).detach().numpy()
        outputs.append(current_predictions)
    output = np.concatenate(outputs, axis=1)
    return output[0, :, :, :]


def load_model(model_scale: float, models_dir: Path = Path("models"), sources: tuple[str] = None) -> tuple[
    UNet, torch.device]:
    """
    Load the model from disk and return it along with the device it's loaded on to.
    :param model_scale: Scale of the model to use. Must be one of [1.0, 0.5, 0.25, 0.125]
    :param models_dir: Path to the directory containing the model files
    :param sources: for testing purposes, specify the sources to download the model from
    :return: Torch model and the device it's loaded on to
    """
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model = UNet(in_channels=1, num_classes=NUM_FLAIR_CLASSES, model_scale=model_scale)
    file_name = get_flair_model_file_name(model_scale)
    model_path = models_dir / file_name

    if not model_path.exists():
        sources = ["Huggingface", "Surfdrive"] if sources is None else sources
        for s in sources:
            if s == "Huggingface":
                try:
                    logger.info(f"Model not found at '{model_path}', downloading from Huggingface")
                    hf_hub_download(repo_id="DroneML/FLAIR-feature-extractor", filename=file_name, local_dir=models_dir)
                except Exception as e:
                    logger.error(f"Failed to download model from Huggingface: {e}")
            if s == "Surfdrive":
                try:
                    logger.info(f"Model not found at '{model_path}', downloading from Surfdrive")
                    surfdrive_file_id = {"flair_toy_ep10_scale1_0.pth": "JzDbL9KWWj5BmtR",
                           "flair_toy_ep15_scale1_0.pth": "zFuHOf3FQBcDzWE"}[file_name]
                    url = f"https://surfdrive.surf.nl/files/index.php/s/{surfdrive_file_id}/download"
                    wget.download(url, out=str(model_path))
                except Exception as e:
                    logger.error(f"Failed to download model from Surfdrive: {e}")
                    raise e
            if model_path.exists():
                logger.info(f"Model successfully downloaded from {s}")
                break

    state = torch.load(model_path, map_location=device, weights_only=True)

    model.load_state_dict(state)
    model.eval()
    return model, device


def pad(input_band: torch.Tensor, band_name):
    """
    Pad the input band, single-sided at the end of width and height axis, to make its dimensions divisible by 16.
    :param input_band: Input band to pad
    :param band_name: Name of the band (for logging)
    :return: Padded input
    """
    width = input_band.shape[2]
    height = input_band.shape[3]
    if width % 16 == 0 and height % 16 == 0:
        padded = input_band
    else:
        pad_left, pad_right = calculate_pad_sizes_1d(width)
        pad_top, pad_bottom = calculate_pad_sizes_1d(height)

        padded = torch.nn.functional.pad(input_band, (pad_top, pad_bottom, pad_left, pad_right))
        logger.info(
            f"Added temporary padding for band {band_name}: (original {height} x {width})"
            f" -> (padded {pad_top + height + pad_bottom} x {pad_left + width + pad_right})"
        )

    return padded


def calculate_pad_sizes_1d(dim_size: int)->tuple[int, int]:
    """"
    Calculate the padding sizes needed to make a dimension size divisible by 16.

    This function computes the amount of padding required before and after the given dimension size
    to make it divisible by 16. The padding is added symmetrically.

    Parameters:
    dim_size (int): The original size of the dimension to be padded.

    Returns:
    tuple[int, int]: A tuple containing the padding size before and after the dimension.
    """
    total_pad = 15 - (dim_size - 1) % 16
    pad_before = total_pad // 2
    pad_after = total_pad - pad_before
    return pad_before, pad_after


def unpad(padded_band:torch.Tensor, original_size):
    """
    Remove padding from the input band to restore its original size.

    This function removes the padding added to the input band to make its dimensions divisible by 16.

    Parameters:
    padded_band (torch.Tensor): The padded input band tensor.
    original_size (tuple[int, int]): The original size of the input band (height, width).

    Returns:
    torch.Tensor: The unpadded input band tensor.
    """
    _,_, original_height, original_width = original_size
    pad_top, pad_bottom = calculate_pad_sizes_1d(original_height)
    pad_left, pad_right = calculate_pad_sizes_1d(original_width)
    return padded_band[:, :, pad_top:pad_top + original_height, pad_left:pad_left + original_width]

def get_flair_model_file_name(model_scale: float) -> str:
    scale_mapping = {1.0: "1_0", 0.5: "0_5", 0.25: "0_25", 0.125: "0_125"}

    scale = None
    for k, v in scale_mapping.items():
        if np.isclose(model_scale, k, atol=0, rtol=0.1):
            scale = v
            break
    if scale is None:
        raise ValueError(f"Unsupported model scale selected ({model_scale}), choose from {scale_mapping.keys()}")

    return f"flair_toy_ep15_scale{scale}.pth"


def get_features_path(raster_path: Path, features_type: FeatureType) -> Path:
    if features_type == FeatureType.IDENTITY:
        return raster_path
    return raster_path.parent / f"{raster_path.stem}_{features_type.name}{raster_path.suffix}"



"""Based on, and parts copied from https://medium.com/@fernandopalominocobo/mastering-u-net-a-step-by-step-guide-to-segmentation-from-scratch-with-pytorch-6a17c5916114"""
import argparse
import time
from pathlib import Path

import torch
from torch import optim, nn
from torch.utils.data import DataLoader, random_split
from torchinfo import summary
from tqdm import tqdm

from segmentmytif.logging_config import setup_logger

try:
    import mlflow

    mlflow_installed = True
except ImportError:
    mlflow_installed = False

from segmentmytif.utils.datasets import MonochromeFlairDataset
from segmentmytif.utils.models import UNet
from segmentmytif.utils.performance_metrics import dice_coefficient


logger = setup_logger(__name__)

def main(root_path, use_mlflow=True, train_set_limit=None, epochs=10, model_scale=1):
    run_name = f"flair_toy_ep{epochs}_scale{str(model_scale).replace('.','_')}{'_tslim_'+str(train_set_limit) if train_set_limit is not None else ''}"

    if use_mlflow and not mlflow_installed:
        raise Exception("Please install mlflow first or specify to run without mlflow.")

    if use_mlflow:
        mlflow.set_tracking_uri(uri="http://127.0.0.1:5000")
        mlflow.set_experiment("Flair feature extractor")


    if use_mlflow:
        mlflow.start_run(run_name=run_name)
        mlflow.set_tag("Training Info", "Segmentation model for ortho photos.")
    train_val_dataset = MonochromeFlairDataset(root_path, limit=train_set_limit)
    test_dataset = MonochromeFlairDataset(root_path, split="test")
    generator = torch.Generator().manual_seed(0)
    train_dataset, validation_dataset = random_split(train_val_dataset, [0.8, 0.2], generator=generator)

    device = "cuda" if torch.cuda.is_available() else "cpu"

    if device == "cuda":
        num_workers = torch.cuda.device_count() * 4
    else:
        num_workers = 1

    LEARNING_RATE = 3e-4
    BATCH_SIZE = 8

    train_dataloader = DataLoader(dataset=train_dataset,
                                  num_workers=num_workers, pin_memory=False,
                                  batch_size=BATCH_SIZE,
                                  shuffle=True)
    validation_dataloader = DataLoader(dataset=validation_dataset,
                                       num_workers=num_workers, pin_memory=False,
                                       batch_size=BATCH_SIZE,
                                       shuffle=True)

    _test_dataloader = DataLoader(dataset=test_dataset,
                                  num_workers=num_workers, pin_memory=False,
                                  batch_size=BATCH_SIZE,
                                  shuffle=True)

    num_classes = 19
    model = UNet(in_channels=1, num_classes=num_classes, model_scale=model_scale).to(device)
    optimizer = optim.AdamW(model.parameters(), lr=LEARNING_RATE)
    criterion = nn.BCEWithLogitsLoss()

    if use_mlflow:
        mlflow.log_param("LEARNING_RATE", LEARNING_RATE)
        mlflow.log_param("BATCH_SIZE", BATCH_SIZE)
        mlflow.log_param("device", device)
        mlflow.log_param("num_workers", num_workers)
        mlflow.log_param("model_scale", model_scale)
        mlflow.log_text(str(summary(model, input_size=(1, 1, 64, 64), verbose=0)), "model_summary.txt")

    torch.cuda.empty_cache()

    train(model, train_dataloader, validation_dataloader, criterion, optimizer, device, epochs, run_name=run_name)

    if use_mlflow:
        mlflow.end_run()


def train(model, train_dataloader, validation_dataloader, criterion, optimizer, device, epochs, run_name):
    for epoch in tqdm(range(epochs)):
        train_loss, train_dice = train_one_step(model, train_dataloader, optimizer, criterion, device)
        val_loss, val_dice = validate(model, validation_dataloader, criterion, device)

        logger.info("-" * 30)
        logger.info(f"Training Loss EPOCH {epoch}: {train_loss:.4f}")
        logger.info(f"Training DICE EPOCH {epoch}: {train_dice:.4f}")
        logger.info("\n")
        logger.info(f"Validation Loss EPOCH {epoch}: {val_loss:.4f}")
        logger.info(f"Validation DICE EPOCH {epoch}: {val_dice:.4f}")
        logger.info("-" * 30)

        metrics = {
            "train_loss": train_loss,
            "train_dc": train_dice,
            "val_loss": val_loss,
            "val_dc": val_dice,
        }

        mlflow.log_metrics(metrics, step=epoch, timestamp=int(round(time.time())))
    # Saving the model
    torch.save(model.state_dict(), f'{run_name}.pth')


def validate(model, validation_dataloader, criterion, device):
    model.eval()
    val_running_loss = 0
    val_running_dc = 0
    with torch.no_grad():
        for idx, img_mask in enumerate(tqdm(validation_dataloader, position=0, leave=True)):
            img = img_mask[0].float().to(device)
            mask = img_mask[1].float().to(device)

            y_pred = model(img)
            loss = criterion(y_pred, mask)
            dc = dice_coefficient(y_pred, mask)

            val_running_loss += loss.item()
            val_running_dc += dc.item()

        val_loss = val_running_loss / (idx + 1)
        val_dc = val_running_dc / (idx + 1)
    return val_dc, val_loss


def train_one_step(model, train_dataloader, optimizer, criterion, device):
    model.train()
    train_running_loss = 0
    train_running_dice = 0
    for idx, img_mask in enumerate(tqdm(train_dataloader, position=0, leave=True)):
        img = img_mask[0].float().to(device)
        mask = img_mask[1].float().to(device)

        y_pred = model(img)
        optimizer.zero_grad()

        dice = dice_coefficient(y_pred, mask)
        loss = criterion(y_pred, mask)

        train_running_loss += loss.item()
        train_running_dice += dice.item()

        loss.backward()
        optimizer.step()
    train_loss = train_running_loss / (idx + 1)
    train_dice = train_running_dice / (idx + 1)
    return train_loss, train_dice,


def parse_args():
    parser = argparse.ArgumentParser(description="Train a semantic segmentation model given a dataset of TIF images.")
    parser.add_argument('-r', '--root', type=Path, required=True, help='Root to the dataset')
    parser.add_argument('--no_mlflow', action='store_true', help='Flag for enabling or disabling MLflow')
    parser.add_argument('--train_set_limit', type=int, default=None,
                        help='Limit for the size of the train set (default: no limit)')
    parser.add_argument('--epochs', type=int, default=10, help='Number of epochs to train (default: 10)')
    parser.add_argument('--model_scale', type=float, default=1,
                        help='Scale number of feature maps in the model (default:1)')
    return parser.parse_args()


if __name__ == '__main__':
    args = parse_args()
    root_path = args.root
    main(root_path, use_mlflow=not args.no_mlflow, train_set_limit=args.train_set_limit, epochs=args.epochs,
         model_scale=args.model_scale)



import logging
from pathlib import Path

import numpy as np
import torch
from PIL import Image
from torch.utils.data import Dataset

from segmentmytif.logging_config import setup_logger
from segmentmytif.utils.io import read_geotiff

logger = setup_logger(__name__)

class MonochromeFlairDataset(Dataset):
    def __init__(self, root_path, limit=None, split="train"):
        self.root_path = root_path
        self.limit = limit
        self.images = sorted([str(p) for p in (Path(root_path) / split / "input").glob("*.tif")])[:self.limit]

        def image_path_to_mask_path(image_path: Path) -> Path:
            return image_path.parent.parent / "labels" / f"MSK{image_path.stem[3:-2]}_0{image_path.suffix}"  # -2 for "_b" where b is band#

        self.masks = [str(image_path_to_mask_path(Path(p))) for p in self.images][:self.limit]
        non_existing_masks = [p for p in self.masks if Path(p).exists() == False]
        if non_existing_masks:
            logging.getLogger().warning(f"{len(non_existing_masks)} of a total of {len(self.masks)} masks not found.")

        if self.limit is None:
            self.limit = len(self.images)

    def __getitem__(self, index):
        img_arr = read_geotiff(self.images[index]).data
        img = normalize_single_band_to_tensor(img_arr[0], debug_note=f"path: {self.images[index]}")[None, :, :]
        mask = load_and_one_hot_encode(self.masks[index])
        return img, mask

    def __len__(self):
        return min(len(self.images), self.limit)

def normalize_single_band_to_tensor(img_arr: np.ndarray, debug_note="") -> torch.Tensor:
    std = img_arr.std()
    if std == 0:
        msg = "Standard deviation = 0 " + debug_note
        logger.debug(msg)
        std = 1
    normalized = (img_arr - img_arr.mean()) / std
    return torch.from_numpy(normalized)


def load_and_one_hot_encode(image_path, num_classes=19):
    """
    Loads a greyscale (labels) image from the specified path and performs one-hot encoding.

    Args:
        image_path (str): The file path to the image.
        num_classes (int, optional): The number of classes for one-hot encoding. Default is 20.

    Returns:
        torch.Tensor: A one-hot encoded tensor with shape (num_classes, height, width).
    """
    image = Image.open(image_path).convert("L")  # Load as grayscale

    image_array = np.array(image, dtype=np.int64) -1  # -1 to convert to zero-based
    image_tensor = torch.from_numpy(image_array)

    one_hot = torch.nn.functional.one_hot(image_tensor, num_classes=num_classes)

    return one_hot.permute(2, 0, 1).float()



import torch


def dice_coefficient(prediction, target, epsilon=1e-07):
    prediction_copy = prediction.clone()

    prediction_copy[prediction_copy < 0] = 0
    prediction_copy[prediction_copy > 0] = 1

    intersection = abs(torch.sum(prediction_copy * target))
    union = abs(torch.sum(prediction_copy) + torch.sum(target))
    dice = (2. * intersection + epsilon) / (union + epsilon)

    return dice
